import ij.*;
import ij.process.*;
import ij.gui.*;

/*
* @author Koutaro Abe ( kabe@cris.hokudai.ac.jp )
* @version 20-April-2010
*/

/**
 * Create a new Stack or a new Slice with a 32bit-Real RAW Stack.
 */
public class Delta {

	private int zD, zS, size;
	private boolean fixFront;
	private ImageStack src;
	private static String cmdEach = "each";
	private static String cmdAll = "total";
	private static String cmdIntegral = "integral";
	
	/**
	 * @param imp
	 *   ImagePlus including source stack.
	 * @param command
	 * 	 'each': difference between each slice and in front of it.
	 *   'total': difference between bottom slice and top slice.
	 *   'integral': difference between each slice and top slice.
	 * @param zDelta
	 *   depth between two slices. ( 0<zDelta )
	 * @param zStep
	 *   increase value of slice counter. ( 0<zStep )
	 */
	public Delta( ImagePlus imp, String command, int zDelta, int zStep ) {
		if( imp==null) {
			IJ.showMessage( "Please select a Stack Image window.");
			return;
		}
		src = imp.getStack();
		if( src!=null && ( ( size=src.getSize() ) > 1) ) {
			if( command.equals( cmdEach ) ) {
				zD = 1; zS = 1; fixFront = false;
			} else if ( command.equals( cmdAll ) ) {
				zD = size-1; zS = size; fixFront = false;
			} else if ( command.equals( cmdIntegral ) ) {
				zD = 1; zS = 1; fixFront = true;
			} else {
				zD = zDelta; zS = zStep; fixFront = false;
				if( zD<1 || zS<1 )
					dialog();
			}
		} else {
			IJ.showMessage( "Selected Image is not a Stack.");
		}
	}
	public Delta( ImagePlus imp, String cmd ) {	this(imp, cmd, 0, 0 ); }
	public Delta( ImagePlus imp ) { this(imp, null, 0, 0 ); }

	private boolean dialog() {
		GenericDialog gd = new GenericDialog( "Delta - Parameters" );
		gd.addNumericField( "Delta between :", 1, 0, 6, " slice(s)" );
		gd.addNumericField( "Interval:", 1, 0, 6, " slice(s)" );
		gd.addCheckbox("Integral", fixFront);	
		gd.showDialog();

		if( gd.wasCanceled() ) return false;
		zD = (int) gd.getNextNumber();
		zS = (int) gd.getNextNumber();
		fixFront = gd.getNextBoolean();
		return true;
	}

	public ImageStack getStack( ) {
		if( src==null || zD==0 || zS==0 ) return null;	
		float[] front, back, result; 
		String frontName, backName;
		int width = (int) src.getWidth();
		int height = (int) src.getHeight();
		int pixels = width * height;	

		IJ.showStatus( "Delta process status" );
		IJ.showProgress( 0.0 );
		ImageStack resultStack = new ImageStack( width, height );
		for( int i = 1; i <= size - zD; i = i + zS  ) {
			IJ.showProgress(  i, size-zD );
			back  = (float[]) src.getPixels( i + zD );
			backName = src.getSliceLabel( i + zD );
			int frontSlice = fixFront ? 1 : i;
			front = (float[]) src.getPixels( frontSlice );
			frontName = src.getSliceLabel( frontSlice );
			ImageProcessor result_ip = new FloatProcessor( width, height );
 			result = (float[]) result_ip.getPixels();
			for( int j = 0; j < pixels; j++ ) 
				result[j] =  back[j] - front[j];
			resultStack.addSlice( frontName + " - " + backName, result_ip );
		}
		return resultStack;
	}
	
	public void show( String title ){
		ImageStack result = getStack();
		if( result==null ) return;
		ImagePlus window = new ImagePlus( title, result );
		if ( window != null ) window.show();
	}

	
}
