import java.awt.*;
import java.awt.event.*;
import java.io.*;

import ij.measure.Calibration;
import ij.measure.Measurements;
import ij.measure.ResultsTable;
import ij.plugin.filter.Analyzer;
import ij.plugin.frame.*;
import ij.*;
import ij.process.*;
import ij.gui.*;
import ij.io.OpenDialog;

/*
* @author Koutaro Abe ( kabe@cris.hokudai.ac.jp )
* @version 20-April-2010
*/

/**
 * Helper plug-in for Isotope Microscope.
*/
public class ImsView_ extends PlugInFrame implements ActionListener {

	Panel panel;
	int previousID;
	static Frame instance;
	public RoiManager rm;

	public ImsView_() {
		super("ImsView");
		if (IJ.versionLessThan("1.39t"))
			return;
		if (instance!=null) {
			instance.toFront();
			return;
		}
		
		instance = this;
		addKeyListener(IJ.getInstance());

		setLayout(new FlowLayout());
		panel = new Panel();
		panel.setLayout(new GridLayout(8, 1, 4, 4));
		addButton("Open");
//		addButton("Delete Current Slice");
		addButton("Total image"); // 1 slice
		addButton("Integral stack");  // n - 0
		addButton("Differential stack");  // n - (n-1)
		addButton("Line");
		addButton("Area");
		addButton("Depth");
		addButton("Sync ROI");
//		addButton("ROI Manager");
//		addButton("Measure All Images");
		add(panel);
		pack();
		GUI.center(this);
		Dimension screen = Toolkit.getDefaultToolkit().getScreenSize();
		instance.setLocation( screen.width - instance.getWidth() - 16 , 16 );
		addKeyListener(IJ.getInstance());

		setVisible(true);
	}
	
	void addButton(String label) {
		Button b = new Button(label);
		b.addActionListener(this);
		b.addKeyListener(IJ.getInstance());
		panel.add(b);
	}

	public void actionPerformed( ActionEvent e ) {
		String label = e.getActionCommand();
		if( label==null )
			return;

		ImagePlus imp = WindowManager.getCurrentImage();
		if( imp!=null ) {
			ImageProcessor ip = imp.getProcessor();
			int id = imp.getID();
			if( id!=previousID )
				ip.snapshot();
			previousID = id;
			new ImsViewRunner( label, imp, ip );
		} else {
			IJ.showStatus( "No image" );
			previousID = 0;
			new ImsViewRunner( label );
		}
	}

	public void processWindowEvent(WindowEvent e) {
		super.processWindowEvent(e);
		if (e.getID()==WindowEvent.WINDOW_CLOSING) {
			instance = null;	
		}
	}
}

class ImsViewRunner extends Thread {
	private String command;
	private ImagePlus imp;
	private ImageProcessor ip;
	public RoiManager rm;

	ImsViewRunner(String command, ImagePlus imp, ImageProcessor ip) {
		super(command);
		this.command = command;
		this.imp = imp;
		this.ip = ip;
		setPriority(Math.max(getPriority()-2, MIN_PRIORITY));
		start();
	}

	ImsViewRunner( String command ) {
		this( command, null, null );
	}

	public void run() {
		try {
			runCommand( command, imp, ip );
		} catch ( OutOfMemoryError e ) {
			IJ.outOfMemory( command );
		} catch( Exception e ) {
			CharArrayWriter caw = new CharArrayWriter( );
			PrintWriter pw = new PrintWriter( caw );
			e.printStackTrace( pw );
			IJ.log( caw.toString( ) );
			IJ.showStatus("");
		}
		if ( imp!=null ) imp.unlock();
	}

	void runCommand(String command, ImagePlus imp, ImageProcessor ip) {
		IJ.showStatus(command + "...");
		if( imp!=null && ip!=null ) {
			if( command.equals( "Delete Current Slice" ) ) { 
				ImageStack s = checkStack(imp);
				if ( s==null )
					return;
				int n = imp.getCurrentSlice();
				s.deleteSlice( n );
				if (s.getSize()==1) {
					imp.setProcessor( null, s.getProcessor(1) );
		 			new ImageWindow( imp );
				}
				imp.setStack( s );
		        imp.unlock();				
			}
			else if ( command.equals("Integral stack") ) {
				if ( !checkImp(imp) )
					return;
				Delta delta = new Delta( imp, "integral" );
				delta.show( "Di-"+imp.getTitle() );
		        imp.unlock();				
			}
			else if ( command.equals("Differential stack") ) {
				if ( !checkImp(imp) )
					return;
				Delta delta = new Delta( imp, "each" );
				delta.show( "Dd-"+imp.getTitle() );
		        imp.unlock();				
			}
			else if ( command.equals("Total image" ) ) {
				if ( !checkImp(imp) )
					return;
				Delta delta = new Delta( imp, "total" );
				delta.show( "Dt-"+imp.getTitle() );
		        imp.unlock();	
			}
			else if( command.equals("Depth") ) {
//				if( checkStack(imp)==null )
//					return;
//				IJ.showMessage("run!");
				IJ.run("Plot Z-axis Profile");
			}
			else if( command.equals("Area") ) {
				IJ.run("Measure");
			}
			else if( command.equals("Line") ) {
				IJ.run("Dynamic Profiler");
			}
			else if ( command.equals("Sync ROI") ) {
				displayRois();
			}
			else if ( command.equals("Measure All Images") ) {
				measureAllImages();
			}
		}

		// 
		if( command.equals("Open") || command.equals("Open Slice") ) {
			boolean stack = false;
			if( command.equals("Open") )
				stack = true;
			try {
				OpenDialog od = new OpenDialog("Open Raw...", null);
				String directory = od.getDirectory();
				String fileName = od.getFileName();
				if (fileName==null)
					return;
				ImportImsDialog d = new ImportImsDialog(fileName, directory, false, stack );
				d.openImage();
			} catch( Exception e ) {
				IJ.showMessage( ""+e );
				return;
			}
			IJ.run("Brightness/Contrast...");
			IJ.run("Enhance Contrast", "saturated=0.35");
		}
		else if ( command.equals("ROI Manager") ) {
			rm = RoiManager.getInstance();
			if( rm==null ){
				rm = new RoiManager();
				rm.toFront();
			}
		}
		if( imp!=null ) {
			imp.updateAndDraw();
			imp.unlock();
		}
		IJ.showStatus("");
	}

	boolean checkImp( ImagePlus imp ) {
		if ( imp==null ) {
			IJ.showMessage( "Please select a image window.");
			return false;
		}
		if ( !imp.lock() ) {
			IJ.showMessage( "Could not lock image.");
			return false;
		}
		return true;
	}

	ImageStack checkStack( ImagePlus imp ){
		if( checkImp( imp )==false )
			return null;
		ImageStack s = imp.getStack();
		if ( s.getSize()<2 ) {
			IJ.showMessage( "Selected image is not a Stack.");
			return null;
		}
		return s;
	}
	void displayRois() {
		Roi roi = imp.getRoi();
		if( roi==null )
			return;
		int[] imageId = WindowManager.getIDList();
		for( int j=0; j<imageId.length; j++ ){
			ImagePlus imp = WindowManager.getImage( imageId[j] );
			if( imp==null )
				continue;
			imp.setRoi( roi, true );
			ImageCanvas ic = imp.getCanvas();
			if( ic==null )
				continue;
			ic.setShowAllROIs(true);
			imp.draw();
		}
	}

	boolean measure( int mode ) {
		ResultsTable rt = new ResultsTable();
		ImagePlus imp = getImage();
		boolean result = measureImp( imp, rt );
		rt.show("Results all images");
		return result;
	}	
	boolean measureAllImages( ) {
		ResultsTable rt = new ResultsTable();
		rt.show("Results all images");
		int[] imageId = WindowManager.getIDList();
		IJ.log( "imageId count:"+imageId.length + ", getImageCount" + WindowManager.getImageCount() );
		for( int i=0; i<imageId.length; i++ ){
			IJ.log( "imageId:"+i+", "+imageId[i] );
			ImagePlus imp = WindowManager.getImage( imageId[i] );
			if( imp==null )
				continue;
			IJ.log( "  run measure" );
			measureImp( imp, rt );
		}
		rt.show("Results all images");
		return true;
	}
	boolean measureImp(ImagePlus imp, ResultsTable rt ) {
		if (imp==null)
			return false;
		int[] indexes = rm.getList().getSelectedIndexes();
		if (indexes.length==0)
			indexes = getAllIndexes();
        if (indexes.length==0) return false;

        boolean allSliceOne = true; // turn false if ROI name contains slice number.
		for( int i=0; i<indexes.length; i++ ) {
			String label = rm.getList().getItem( indexes[i] );
			if( rm.getSliceNumber(label)>1 )
				allSliceOne = false;
//			Roi roi = (Roi) rm.getROIs().get(label);
		}
		
		int measurements = Analyzer.getMeasurements();
		if( imp.getStackSize()>1 )
			Analyzer.setMeasurements(measurements|Measurements.SLICE);
		
		int currentSlice = imp.getCurrentSlice();
		for( int i=0; i<indexes.length; i++ ) {
			if( restore( imp, indexes[i], !allSliceOne ) ) {
				Analyzer analyzer = new Analyzer(imp, measurements, rt);
				analyzer.measure();
//				IJ.run( "Measure" );
			} else
				break;
		}

		imp.setSlice( currentSlice );
		Analyzer.setMeasurements( measurements );
		if (indexes.length>1)
			IJ.run("Select None");
		return true;
	}	

	int[] getAllIndexes() {
		int count = rm.getList().getItemCount();
		int[] indexes = new int[count];
		for (int i=0; i<count; i++)
			indexes[i] = i;
		return indexes;
	}
	ImagePlus getImage() {
		ImagePlus imp = WindowManager.getCurrentImage();
		if (imp==null) {
			error("There are no images open.");
			return null;
		} else
			return imp;
	}
	ImagePlus getImage( int id ) {
		ImagePlus imp = WindowManager.getImage( id );
		if (imp==null) {
			error("There are no images open.");
			return null;
		} else
			return imp;
	}

	boolean restore(ImagePlus imp, int index, boolean setSlice) {
		String label = rm.getList().getItem(index);
		Roi roi = (Roi)rm.getROIs().get(label);
		if (imp==null || roi==null)
			return false;
        if (setSlice) {
            int n = rm.getSliceNumber(label);
            if (n>=1 && n<=imp.getStackSize()) {
            	if (imp.isHyperStack()||imp.isComposite())
                	imp.setPosition(n);
                else
                	imp.setSlice(n);
            }
        }
        Roi roi2 = (Roi)roi.clone();
		Calibration cal = imp.getCalibration();
		Rectangle r = roi2.getBounds();
		if (cal.xOrigin!=0.0 || cal.yOrigin!=0.0)
			roi2.setLocation(r.x+(int)cal.xOrigin, r.y+(int)cal.yOrigin);
		int width= imp.getWidth(), height=imp.getHeight();
/*		if (restoreCentered) {
			ImageCanvas ic = imp.getCanvas();
			if (ic!=null) {
				Rectangle r1 = ic.getSrcRect();
				Rectangle r2 = roi2.getBounds();
				roi2.setLocation(r1.x+r1.width/2-r2.width/2, r1.y+r1.height/2-r2.height/2);
			}
		}
*/		if (r.x>=width || r.y>=height || (r.x+r.width)<=0 || (r.y+r.height)<=0)
			roi2.setLocation((width-r.width)/2, (height-r.height)/2);
//		if (noUpdateMode) {
//			imp.setRoi(roi2, false);
//			noUpdateMode = false;
//		} else
			imp.setRoi(roi2, true);
		return true;
	}

	boolean error(String msg) {
		IJ.showMessage(msg);
		Macro.abort();
		return false;
	}
}
